"""commandline interactions

Future versions could expand this into a full commandline interface
for scripting spectrometer things.

  TODO
    - implement main() to be used as an entry point
    - handle cmdline with argparse
    - basic spectrum acquisition should be enough
    - has to be stateless! open/close overhead is acceptable

WARNING
-------
This submodule is not stable. Anything in here can change even between
micro version releases. Do not rely on this yet. If you want this to
be a public interface open an issue on GitHub and/or work on a Pull
Request implementing the functionality.

author: Andreas Poehlmann
"""
from __future__ import print_function

import importlib
import itertools
import operator


def ls():
    """INTERNAL ONLY: print connected spectrometers"""
    connected = []
    for backend in ("cseabreeze", "pyseabreeze"):
        # noinspection PyBroadException
        try:
            sb_backend = importlib.import_module("seabreeze.{}".format(backend))
            api = sb_backend.SeaBreezeAPI()
            try:
                devices = api.list_devices()
                for d in devices:
                    # noinspection PyProtectedMember
                    connected.append((d.model, d.serial_number, sb_backend._backend_))
            finally:
                api.shutdown()
        except Exception:
            pass
        finally:
            # noinspection PyUnusedLocal
            sb_backend = None

    key_func = operator.itemgetter(0, 1)  # sort by (model, serial)
    connected = sorted(connected, key=key_func)

    # output format:
    # MY_MODEL   SERIAL_1234  cseabreeze,pyseabreeze
    for model_serial_number, info in itertools.groupby(connected, key=key_func):
        model, serial = model_serial_number
        print("{}\t{}\t{}".format(model, serial, ",".join(b[2] for b in info)))



"""custom seabreeze exceptions"""
# DO NOT DIRECTLY IMPORT EXCEPTIONS FROM HERE!
# ALWAYS IMPORT FROM `seabreeze.spectrometers`


class SeaBreezeError(Exception):
    """SeaBreezeError base class"""
    pass


class SeaBreezeNumFeaturesError(SeaBreezeError):
    pass


class SeaBreezeNotSupported(SeaBreezeError):
    pass



import struct
import warnings

import numpy

from seabreeze.pyseabreeze.exceptions import SeaBreezeError, SeaBreezeNotSupported
from seabreeze.pyseabreeze.features._base import SeaBreezeFeature
from seabreeze.pyseabreeze.features.eeprom import SeaBreezeEEPromFeatureOOI
from seabreeze.pyseabreeze.protocol import OBPProtocol, OOIProtocol


# Definition
# ==========
#
class SeaBreezeSpectrometerFeature(SeaBreezeFeature):
    identifier = "spectrometer"

    def set_trigger_mode(self, mode):
        raise NotImplementedError("implement in derived class")

    def set_integration_time_micros(self, integration_time_micros):
        raise NotImplementedError("implement in derived class")

    def get_integration_time_micros_limits(self):
        raise NotImplementedError("implement in derived class")

    def get_maximum_intensity(self):
        raise NotImplementedError("implement in derived class")

    def get_electric_dark_pixel_indices(self):
        raise NotImplementedError("implement in derived class")

    def _spectrum_length(self):
        raise NotImplementedError("implement in derived class")

    def get_wavelengths(self):
        raise NotImplementedError("implement in derived class")

    def get_intensities(self):
        raise NotImplementedError("implement in derived class")

    def _get_spectrum_raw(self):
        raise NotImplementedError("implement in derived class")

    def get_fast_buffer_spectrum(self):
        raise SeaBreezeNotSupported(
            "needs to be provided in the specific implementation if supported"
        )


# Spectrometer Features based on USBCommOOI
# =========================================
#
class SeaBreezeSpectrometerFeatureOOI(SeaBreezeSpectrometerFeature):
    _required_protocol_cls = OOIProtocol
    _required_features = ("eeprom",)
    _required_kwargs = (
        "dark_pixel_indices",
        "integration_time_min",
        "integration_time_max",
        "integration_time_base",
        "spectrum_num_pixel",
        "spectrum_raw_length",
        "spectrum_max_value",
        "trigger_modes",
    )

    _normalization_value = 1.0

    # config
    _dark_pixel_indices = None
    _integration_time_min = None
    _integration_time_max = None
    _integration_time_base = None
    _spectrum_num_pixel = None
    _spectrum_raw_length = None
    _spectrum_max_value = None
    _trigger_modes = None

    def __init__(self, device, feature_id, **kwargs):
        super(SeaBreezeSpectrometerFeatureOOI, self).__init__(
            device, feature_id, **kwargs
        )
        self._dark_pixel_indices = kwargs["dark_pixel_indices"]
        self._integration_time_min = kwargs["integration_time_min"]
        self._integration_time_max = kwargs["integration_time_max"]
        self._integration_time_base = kwargs["integration_time_base"]
        self._spectrum_num_pixel = kwargs["spectrum_num_pixel"]
        self._spectrum_raw_length = kwargs["spectrum_raw_length"]
        self._spectrum_max_value = kwargs["spectrum_max_value"]
        self._trigger_modes = kwargs["trigger_modes"]

    def set_trigger_mode(self, mode):
        if mode in self._trigger_modes:
            self.protocol.send(0x0A, mode)
        else:
            raise SeaBreezeError("Only supports: %s" % str(self._trigger_modes))

    def set_integration_time_micros(self, integration_time_micros):
        t_min = self._integration_time_min
        t_max = self._integration_time_max
        if t_min <= integration_time_micros < t_max:
            i_time = int(integration_time_micros / self._integration_time_base)
            self.protocol.send(0x02, i_time)
        else:
            raise SeaBreezeError("Integration not in [{:d}, {:d}]".format(t_min, t_max))

    def get_integration_time_micros_limits(self):
        return self._integration_time_min, self._integration_time_max

    def get_maximum_intensity(self):
        return float(self._spectrum_max_value)

    def get_electric_dark_pixel_indices(self):
        return list(self._dark_pixel_indices)

    @property
    def _spectrum_length(self):
        return self._spectrum_num_pixel

    def get_wavelengths(self):
        indices = numpy.arange(self._spectrum_length, dtype=numpy.float64)
        # OOI spectrometers store the wavelength calibration in slots 1,2,3,4
        coeffs = []
        for i in range(1, 5):
            # noinspection PyProtectedMember
            coeffs.append(
                float(
                    SeaBreezeEEPromFeatureOOI._func_eeprom_read_slot(self.protocol, i)
                )
            )
        return sum(wl * (indices ** i) for i, wl in enumerate(coeffs))

    def get_intensities(self):
        tmp = self._get_spectrum_raw()
        ret = numpy.array(
            struct.unpack("<" + "H" * self._spectrum_length, tmp[:-1]),
            dtype=numpy.double,
        )
        return ret * self._normalization_value

    def _get_spectrum_raw(self):
        tmp = numpy.empty((self._spectrum_raw_length,), dtype=numpy.uint8)
        self.protocol.send(0x09)

        timeout = int(
            self._integration_time_max * 1e-3
            + self.protocol.transport.default_timeout_ms
        )
        # noinspection PyProtectedMember
        tmp[:] = self.protocol.receive(
            size=self._spectrum_raw_length,
            timeout_ms=timeout,
            mode=self.protocol.transport._default_read_spectrum_endpoint,
        )
        return tmp

    def get_fast_buffer_spectrum(self):
        raise SeaBreezeNotSupported(
            "needs to be provided in the specific implementation if supported"
        )


class SeaBreezeSpectrometerFeatureOOI2K(SeaBreezeSpectrometerFeatureOOI):
    def get_intensities(self):
        tmp = self._get_spectrum_raw()
        # The byte order is different for some models
        N_raw = self._spectrum_raw_length - 1
        N_pix = self._spectrum_length
        idx = [(i // 2) % 64 + (i % 2) * 64 + (i // 128) * 128 for i in range(N_raw)]
        # high nibble not guaranteed to be pulled low
        tsorted = tmp[idx] & numpy.array((0xFF, 0x0F) * N_pix, dtype=numpy.uint8)
        ret = numpy.array(struct.unpack("<" + "H" * N_pix, tsorted), dtype=numpy.double)
        # sorted and parsed
        return ret * self._normalization_value


class SeaBreezeSpectrometerFeatureOOIFPGA(SeaBreezeSpectrometerFeatureOOI):
    def __init__(self, device, feature_id, **kwargs):
        super(SeaBreezeSpectrometerFeatureOOIFPGA, self).__init__(
            device, feature_id, **kwargs
        )
        self.protocol.send(0xFE)
        ret = self.protocol.receive(size=16)
        data = struct.unpack("<HLBBBBBBBBBB", ret[:])
        speed = data[10]
        if speed == 0x00:
            # USB1.0 port
            self.protocol.transport._default_read_spectrum_endpoint = "low_speed"
            warnings.warn(
                "The device is connected to a USB1.0 port. "
                "This might not work. Please report this issue upstream."
            )
        elif speed == 0x80:
            # USB2.0+ port
            self.protocol.transport._default_read_spectrum_endpoint = "high_speed"
        else:
            # raise if the value is not what's expected from the datasheets
            raise ValueError(
                "Unknown speed setting '{}' please file this as a bug!".format(speed)
            )


class SeaBreezeSpectrometerFeatureOOIFPGA4K(SeaBreezeSpectrometerFeatureOOIFPGA):
    def _get_spectrum_raw(self):
        tmp = numpy.empty((self._spectrum_raw_length,), dtype=numpy.uint8)
        timeout = int(
            self._integration_time_max * 1e-3
            + self.protocol.transport.default_timeout_ms
        )
        self.protocol.send(0x09)
        # noinspection PyProtectedMember
        if self.protocol.transport._default_read_spectrum_endpoint == "low_speed":
            tmp[:] = self.protocol.receive(
                size=self._spectrum_raw_length, timeout_ms=timeout
            )
        else:  # high_speed
            tmp[:2048] = self.protocol.receive(
                size=2048, timeout_ms=timeout, mode="high_speed_alt"
            )
            tmp[2048:] = self.protocol.receive(
                size=self._spectrum_raw_length - 2048,
                timeout_ms=timeout,
                mode="high_speed",
            )
        return tmp


class _SeaBreezeSpectrometerSaturationMixin(object):
    def _saturation_unpack(self, ret):
        return struct.unpack("<H", ret[6:8])[0]

    def _saturation_not_initialized(self, x):
        return x == 0

    # noinspection PyUnresolvedReferences
    def _saturation_get_normalization_value(self):
        """internal only"""
        # noinspection PyProtectedMember
        ret = SeaBreezeEEPromFeatureOOI._func_eeprom_read_slot(
            self.protocol, 17, raw=True
        )
        # ret contains the first two response bytes, then the eeprom data
        saturation = self._saturation_unpack(ret)
        if self._saturation_not_initialized(saturation):
            # pass  # not initialized?
            return self._normalization_value
        else:
            return float(self._spectrum_max_value) / saturation


class SeaBreezeSpectrometerFeatureOOIGain(
    SeaBreezeSpectrometerFeatureOOI, _SeaBreezeSpectrometerSaturationMixin
):
    def __init__(self, device, feature_id, **kwargs):
        # set the usbspeed
        super(SeaBreezeSpectrometerFeatureOOIGain, self).__init__(
            device, feature_id, **kwargs
        )
        # load the saturation value
        self._normalization_value = self._saturation_get_normalization_value()


class SeaBreezeSpectrometerFeatureOOIFPGAGain(
    SeaBreezeSpectrometerFeatureOOIFPGA, _SeaBreezeSpectrometerSaturationMixin
):
    def __init__(self, device, feature_id, **kwargs):
        # set the usbspeed
        super(SeaBreezeSpectrometerFeatureOOIFPGAGain, self).__init__(
            device, feature_id, **kwargs
        )
        # load the saturation value
        self._normalization_value = self._saturation_get_normalization_value()


class SeaBreezeSpectrometerFeatureOOIFPGA4KGain(
    SeaBreezeSpectrometerFeatureOOIFPGA4K, _SeaBreezeSpectrometerSaturationMixin
):
    def __init__(self, device, feature_id, **kwargs):
        # set the usbspeed
        super(SeaBreezeSpectrometerFeatureOOIFPGA4KGain, self).__init__(
            device, feature_id, **kwargs
        )
        # get the saturation value
        self._normalization_value = self._saturation_get_normalization_value()


class SeaBreezeSpectrometerFeatureOOIGainAlt(SeaBreezeSpectrometerFeatureOOIGain):
    # XXX: The NIRQUEST stores this value somewhere else
    #      and might also not have been programmed yet..
    # TODO: And this is planned for the QE65000 apparently.
    def _saturation_unpack(self, ret):
        return struct.unpack("<L", ret[6:10])[0]


class SeaBreezeSpectrometerFeatureOOIFPGAGainAlt(
    SeaBreezeSpectrometerFeatureOOIFPGAGain
):
    # XXX: The Apex, Maya2000pro, MayaLSL store this value somewhere else
    #      and might also not have been programmed yet...
    # ret contains the first two response bytes, then the eeprom data
    def _saturation_unpack(self, ret):
        return struct.unpack("<H", ret[2:4])[0]

    def _saturation_not_initialized(self, x):
        return x <= 32768 or x > self._spectrum_max_value


# Spectrometer Features based on USBCommOBP
# =========================================
#
class SeaBreezeSpectrometerFeatureOBP(SeaBreezeSpectrometerFeature):
    _required_protocol_cls = OBPProtocol
    # required_interface_cls = USBCommOBP
    required_features = ()
    _required_kwargs = (
        "dark_pixel_indices",
        "integration_time_min",
        "integration_time_max",
        "integration_time_base",
        "spectrum_num_pixel",
        "spectrum_raw_length",
        "spectrum_max_value",
        "trigger_modes",
    )

    _normalization_value = 1.0

    # config
    _dark_pixel_indices = None
    _integration_time_min = None
    _integration_time_max = None
    _integration_time_base = None
    _spectrum_num_pixel = None
    _spectrum_raw_length = None
    _spectrum_max_value = None
    _trigger_modes = None

    def __init__(self, device, feature_id, **kwargs):
        super(SeaBreezeSpectrometerFeatureOBP, self).__init__(
            device, feature_id, **kwargs
        )
        self._dark_pixel_indices = kwargs["dark_pixel_indices"]
        self._integration_time_min = kwargs["integration_time_min"]
        self._integration_time_max = kwargs["integration_time_max"]
        self._integration_time_base = kwargs["integration_time_base"]
        self._spectrum_num_pixel = kwargs["spectrum_num_pixel"]
        self._spectrum_raw_length = kwargs["spectrum_raw_length"]
        self._spectrum_max_value = kwargs["spectrum_max_value"]
        self._trigger_modes = kwargs["trigger_modes"]

    def set_trigger_mode(self, mode):
        if mode in self._trigger_modes:
            self.protocol.send(0x00110110, mode, request_ack=True)
        else:
            raise SeaBreezeError("Only supports: %s" % str(self._trigger_modes))

    def set_integration_time_micros(self, integration_time_micros):
        t_min = self._integration_time_min
        t_max = self._integration_time_max
        if t_min <= integration_time_micros < t_max:
            i_time = int(integration_time_micros / self._integration_time_base)
            self.protocol.send(0x00110010, i_time)
        else:
            raise SeaBreezeError("Integration not in [{:d}, {:d}]".format(t_min, t_max))

    def get_integration_time_micros_limits(self):
        return self._integration_time_min, self._integration_time_max

    def get_maximum_intensity(self):
        return float(self._spectrum_max_value)

    def get_electric_dark_pixel_indices(self):
        return list(self._dark_pixel_indices)

    @property
    def _spectrum_length(self):
        return self._spectrum_num_pixel

    def get_wavelengths(self):
        # get number of wavelength coefficients
        data = self.protocol.query(0x00180100)
        N = struct.unpack("<B", data)[0]
        # now query the coefficients
        coeffs = []
        for i in range(N):
            data = self.protocol.query(0x00180101, i)
            coeffs.append(struct.unpack("<f", data)[0])
        # and generate the wavelength array
        indices = numpy.arange(self._spectrum_length, dtype=numpy.float64)
        return sum(wl * (indices ** i) for i, wl in enumerate(coeffs))

    def get_intensities(self):
        tmp = self._get_spectrum_raw()
        # noinspection PyTypeChecker
        return numpy.array(
            struct.unpack("<" + "H" * self._spectrum_length, tmp), dtype=numpy.double
        )

    def _get_spectrum_raw(self):
        timeout = int(
            self._integration_time_max * 1e-3
            + self.protocol.transport.default_timeout_ms
        )
        datastring = self.protocol.query(0x00101100, timeout_ms=timeout)
        return numpy.frombuffer(datastring, dtype=numpy.uint8)

    def get_fast_buffer_spectrum(self):
        raise SeaBreezeNotSupported(
            "needs to be provided in the specific implementation if supported"
        )


# Model specific changes
# ======================
#
class SeaBreezeSpectrometerFeatureUSB2000(SeaBreezeSpectrometerFeatureOOI2K):
    pass


class SeaBreezeSpectrometerFeatureHR2000(SeaBreezeSpectrometerFeatureOOI2K):
    pass


class SeaBreezeSpectrometerFeatureHR4000(SeaBreezeSpectrometerFeatureOOIFPGA4K):
    def get_intensities(self):
        tmp = self._get_spectrum_raw()
        # The HR4000 needs to xor with 0x2000
        ret = (
            numpy.array(struct.unpack("<" + "H" * self._spectrum_length, tmp[:-1]))
            ^ 0x2000
        )
        return ret.astype(numpy.double) * self._normalization_value


class SeaBreezeSpectrometerFeatureUSB650(SeaBreezeSpectrometerFeatureOOI2K):
    pass


class SeaBreezeSpectrometerFeatureHR2000PLUS(SeaBreezeSpectrometerFeatureOOIFPGA):
    def get_intensities(self):
        tmp = self._get_spectrum_raw()
        # The HR2000PLUS needs to xor with 0x2000
        ret = (
            numpy.array(struct.unpack("<" + "H" * self._spectrum_length, tmp[:-1]))
            ^ 0x2000
        )
        return ret.astype(numpy.double) * self._normalization_value


class SeaBreezeSpectrometerFeatureQE65000(SeaBreezeSpectrometerFeatureOOIFPGA):
    def get_wavelengths(self):
        # QE65000 specific override
        indices = numpy.arange(-10, self._spectrum_length - 10, dtype=numpy.float64)
        # OOI spectrometers store the wavelength calibration in slots 1,2,3,4
        coeffs = []
        for i in range(1, 5):
            # noinspection PyProtectedMember
            coeffs.append(
                float(
                    SeaBreezeEEPromFeatureOOI._func_eeprom_read_slot(self.protocol, i)
                )
            )
        return sum(wl * (indices ** i) for i, wl in enumerate(coeffs))

    def get_intensities(self):
        tmp = self._get_spectrum_raw()
        # The QE65000 needs to xor with 0x8000
        ret = (
            numpy.array(struct.unpack("<" + "H" * self._spectrum_length, tmp[:-1]))
            ^ 0x8000
        )
        return ret.astype(numpy.double) * self._normalization_value


class SeaBreezeSpectrometerFeatureUSB2000PLUS(SeaBreezeSpectrometerFeatureOOIFPGAGain):
    pass


class SeaBreezeSpectrometerFeatureUSB4000(SeaBreezeSpectrometerFeatureOOIFPGA4KGain):
    pass


class SeaBreezeSpectrometerFeatureNIRQUEST512(SeaBreezeSpectrometerFeatureOOIGainAlt):
    def get_intensities(self):
        tmp = self._get_spectrum_raw()
        # The NIRQUEST512 needs to xor with 0x8000
        ret = (
            numpy.array(struct.unpack("<" + "H" * self._spectrum_length, tmp[:-1]))
            ^ 0x8000
        )
        return ret.astype(numpy.double) * self._normalization_value


class SeaBreezeSpectrometerFeatureNIRQUEST256(SeaBreezeSpectrometerFeatureOOIGainAlt):
    def get_intensities(self):
        tmp = self._get_spectrum_raw()
        # The NIRQUEST256 needs to xor with 0x8000
        ret = (
            numpy.array(struct.unpack("<" + "H" * self._spectrum_length, tmp[:-1]))
            ^ 0x8000
        )
        return ret.astype(numpy.double) * self._normalization_value


class SeaBreezeSpectrometerFeatureMAYA2000PRO(
    SeaBreezeSpectrometerFeatureOOIFPGAGainAlt
):
    pass


class SeaBreezeSpectrometerFeatureMAYA2000(SeaBreezeSpectrometerFeatureOOIFPGA):
    pass


class SeaBreezeSpectrometerFeatureTORUS(SeaBreezeSpectrometerFeatureOOIFPGAGain):
    # The Torus uses the USB2000Plus spec feature
    pass


class SeaBreezeSpectrometerFeatureAPEX(SeaBreezeSpectrometerFeatureOOIFPGAGainAlt):
    pass


class SeaBreezeSpectrometerFeatureMAYALSL(SeaBreezeSpectrometerFeatureOOIFPGAGainAlt):
    pass


class SeaBreezeSpectrometerFeatureJAZ(SeaBreezeSpectrometerFeatureOOIGain):
    def get_intensities(self):
        tmp = self._get_spectrum_raw()
        # XXX: No sync byte for the Jaz
        ret = numpy.array(
            struct.unpack("<" + "H" * self._spectrum_length, tmp[:]), dtype=numpy.double
        )
        return ret * self._normalization_value


class SeaBreezeSpectrometerFeatureSTS(SeaBreezeSpectrometerFeatureOBP):
    pass


class SeaBreezeSpectrometerFeatureQEPRO(SeaBreezeSpectrometerFeatureOBP):
    def _get_spectrum_raw(self):
        timeout = int(
            self._integration_time_max * 1e-3
            + self.protocol.transport.default_timeout_ms
        )
        datastring = self.protocol.query(0x00100928, timeout_ms=timeout)
        return numpy.fromstring(datastring, dtype=numpy.uint8)

    def get_intensities(self):
        tmp = self._get_spectrum_raw()
        # 32byte metadata block at beginning
        ret = numpy.array(
            struct.unpack("<" + "I" * self._spectrum_length, tmp[32:]),
            dtype=numpy.double,
        )
        return ret * self._normalization_value


class SeaBreezeSpectrometerFeatureVENTANA(SeaBreezeSpectrometerFeatureOBP):
    pass


class SeaBreezeSpectrometerFeatureSPARK(SeaBreezeSpectrometerFeatureOBP):
    pass


class SeaBreezeSpectrometerFeatureHDX(SeaBreezeSpectrometerFeatureOBP):
    def _get_spectrum_raw(self):
        timeout = int(
            self._integration_time_max * 1e-3
            + self.protocol.transport.default_timeout_ms
        )
        # the message type is different than the default defined in the protocol,
        # requires addition of a new message type in protocol to work
        datastring = self.protocol.query(0x00101000, timeout_ms=timeout)
        return numpy.fromstring(datastring, dtype=numpy.uint8)



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeGPIOFeature(SeaBreezeFeature):
    identifier = "gpio"

    modes = dict(
        GPIO_OUTPUT_PUSH_PULL=0x00,
        GPIO_OPEN_DRAIN_OUTPUT=0x01,
        DAC_OUTPUT=0x02,
        GPIO_INPUT_HIGH_Z=0x80,
        GPIO_INPUT_PULL_DOWN=0x81,
        ADC_INPUT=0x82,
    )

    def get_number_of_gpio_pins(self):
        raise NotImplementedError("implement in derived class")

    def get_gpio_output_enable_vector(self):
        raise NotImplementedError("implement in derived class")

    def set_gpio_output_enable_vector(self, output_enable_vector, bit_mask):
        raise NotImplementedError("implement in derived class")

    def get_gpio_value_vector(self):
        raise NotImplementedError("implement in derived class")

    def set_gpio_value_vector(self, value_vector, bit_mask):
        raise NotImplementedError("implement in derived class")

    def get_number_of_egpio_pins(self):
        raise NotImplementedError("implement in derived class")

    def get_egpio_available_modes(self, pin_number):
        raise NotImplementedError("implement in derived class")

    def get_egpio_current_mode(self, pin_number):
        raise NotImplementedError("implement in derived class")

    def set_egpio_mode(self, pin_number, mode, value=0.0):
        raise NotImplementedError("implement in derived class")

    def get_egpio_output_vector_vector(self):
        raise NotImplementedError("implement in derived class")

    def set_egpio_output_vector(self, output_vector, bit_mask):
        raise NotImplementedError("implement in derived class")

    def get_egpio_value(self, pin_number):
        raise NotImplementedError("implement in derived class")

    def set_egpio_value(self, pin_number, value):
        raise NotImplementedError("implement in derived class")



import struct

from seabreeze.pyseabreeze.features._base import SeaBreezeFeature
from seabreeze.pyseabreeze.features.eeprom import SeaBreezeEEPromFeatureOOI
from seabreeze.pyseabreeze.protocol import OBPProtocol, OOIProtocol


# Definition
# ==========
#
class SeaBreezeNonlinearityCoefficientsFeature(SeaBreezeFeature):
    identifier = "nonlinearity_coefficients"

    def get_nonlinearity_coefficients(self):
        raise NotImplementedError("implement in derived class")


# OOI implementation
# ==================
#
class NonlinearityCoefficientsEEPromFeatureOOI(
    SeaBreezeNonlinearityCoefficientsFeature
):
    _required_protocol_cls = OOIProtocol
    _required_features = ("eeprom",)

    def get_nonlinearity_coefficients(self):
        # The spectrometers store the wavelength calibration in slots 6..13
        coeffs = []
        # noinspection PyProtectedMember
        order = int(
            float(SeaBreezeEEPromFeatureOOI._func_eeprom_read_slot(self.protocol, 14))
        )
        for i in range(6, 6 + order + 1):
            # noinspection PyProtectedMember
            coeffs.append(
                float(
                    SeaBreezeEEPromFeatureOOI._func_eeprom_read_slot(self.protocol, i)
                )
            )
        return coeffs


# OBP implementation
# ==================
#
class NonlinearityCoefficientsFeatureOBP(SeaBreezeNonlinearityCoefficientsFeature):
    _required_protocol_cls = OBPProtocol

    def get_nonlinearity_coefficients(self):
        # get number of nonlinearity coefficients
        data = self.protocol.query(0x00181100)
        N = struct.unpack("<B", data)[0]
        # now query the coefficients
        coeffs = []
        for i in range(N):
            data = self.protocol.query(0x00181101, i)
            coeffs.append(struct.unpack("<f", data)[0])
        return coeffs



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeDataBufferFeature(SeaBreezeFeature):
    identifier = "data_buffer"

    def clear(self):
        raise NotImplementedError("implement in derived class")

    def remove_oldest_spectra(self, number_of_spectra):
        raise NotImplementedError("implement in derived class")

    def get_number_of_elements(self):
        raise NotImplementedError("implement in derived class")

    def get_buffer_capacity(self):
        raise NotImplementedError("implement in derived class")

    def set_buffer_capacity(self, capacity):
        raise NotImplementedError("implement in derived class")

    def get_buffer_capacity_maximum(self):
        raise NotImplementedError("implement in derived class")

    def get_buffer_capacity_minimum(self):
        raise NotImplementedError("implement in derived class")



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeIrradCalFeature(SeaBreezeFeature):
    identifier = "irrad_cal"

    def has_collection_area(self):
        raise NotImplementedError("implement in derived class")

    def read_collection_area(self):
        raise NotImplementedError("implement in derived class")

    def write_collection_area(self, area):
        raise NotImplementedError("implement in derived class")

    def read_calibration(self):
        raise NotImplementedError("implement in derived class")

    def write_calibration(self, calibration_array):
        raise NotImplementedError("implement in derived class")



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeAcquisitionDelayFeature(SeaBreezeFeature):
    identifier = "acquisition_delay"

    def set_delay_microseconds(self, delay_usec):
        raise NotImplementedError("implement in derived class")

    def get_delay_microseconds(self):
        raise NotImplementedError("implement in derived class")

    delay_microseconds = property(get_delay_microseconds, set_delay_microseconds)

    def get_delay_increment_microseconds(self):
        raise NotImplementedError("implement in derived class")

    def get_maximum_delay_microseconds(self):
        raise NotImplementedError("implement in derived class")

    def get_minimum_delay_microseconds(self):
        raise NotImplementedError("implement in derived class")



import struct
import time
import enum


class _FPGARegisterFeatureOOI(object):
    """extract fpga register feature

    (has no equivalent in cseabreeze)
    """

    class Codes(enum.IntEnum):
        FIRMWARE_VERSION = 0x04
        V1_CONTINUOUS_STROBE_TIMER_INTERVAL_DIVISOR = (
            V3_CONTINUOUS_STROBE_TIMER_MSB
        ) = 0x08
        V1_CONTINUOUS_STROBE_BASE_CLOCK_DIVISOR = V3_CONTINUOUS_STROBE_TIMER_LSB = 0x0C

    def __init__(self, protocol):
        self.protocol = protocol

    def read_register(self, register):
        fw_raw = self.protocol.query(0x6B, register)
        command, data = struct.unpack("<BH", fw_raw)
        assert command in {
            self.Codes.FIRMWARE_VERSION,  # as specified in datasheets
            0xFE  # FlameS seems to return this?
        }, str(list(map(hex, fw_raw)))
        return data

    def write_register(self, register, data):
        self.protocol.send(0x6A, payload=(register, data))
        time.sleep(0.0001)  # guarantee 100us sleep between commands

    def get_firmware_version(self):
        data = self.read_register(self.Codes.FIRMWARE_VERSION)
        fw_version = (
            (data >> 12) & 0x0F,  # major
            (data >> 4) & 0xFF,  # minor
            data & 0xF,  # patch
        )
        return fw_version



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeI2CMasterFeature(SeaBreezeFeature):
    identifier = "i2c_master"

    def get_number_of_buses(self):
        raise NotImplementedError("implement in derived class")

    def read_bus(self, bus_index, slave_address, buffer_length=1024):
        raise NotImplementedError("implement in derived class")

    def write_bus(self, bus_index, slave_address, data):
        raise NotImplementedError("implement in derived class")



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeLightSourceFeature(SeaBreezeFeature):
    identifier = "light_source"

    def get_count(self):
        raise NotImplementedError("implement in derived class")

    def has_enable(self, light_source_index):
        raise NotImplementedError("implement in derived class")

    def is_enabled(self, light_source_index):
        raise NotImplementedError("implement in derived class")

    def set_enable(self, light_source_index, enable):
        raise NotImplementedError("implement in derived class")

    def has_variable_intensity(self, light_source_index):
        raise NotImplementedError("implement in derived class")

    def get_intensity(self, light_source_index):
        raise NotImplementedError("implement in derived class")

    def set_intensity(self, light_source_index, intensity):
        raise NotImplementedError("implement in derived class")



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeOpticalBenchFeature(SeaBreezeFeature):
    identifier = "optical_bench"

    def get_fiber_diameter_microns(self):
        raise NotImplementedError("implement in derived class")

    def get_slit_width_microns(self):
        raise NotImplementedError("implement in derived class")

    def get_id(self):
        raise NotImplementedError("implement in derived class")

    def get_serial_number(self):
        raise NotImplementedError("implement in derived class")

    def get_coating(self):
        raise NotImplementedError("implement in derived class")

    def get_filter(self):
        raise NotImplementedError("implement in derived class")

    def get_grating(self):
        raise NotImplementedError("implement in derived class")



"""Implementation of the Protocol layer

ocean optics spectrometers use two different communication protocols:
- one is called 'OBP protocol'  # OceanBinaryProtocol
- the other one 'OOI protocol'  # ??? OceanOpticsInterface ??? maybe

"""
import functools
import hashlib
import struct
import time
import warnings
import weakref

from seabreeze.pyseabreeze.exceptions import SeaBreezeError

try:
    from collections.abc import Callable
except ImportError:  # Python27
    from collections import Callable


class ProtocolInterface(object):
    def __init__(self, transport):
        if not (
            hasattr(transport, "write")
            and isinstance(transport.write, Callable)
            and hasattr(transport, "read")
            and isinstance(transport.read, Callable)
        ):
            raise TypeError("transport does not implement read and write methods")
        self.transport = weakref.proxy(transport)

    def send(self, msg_type, payload, timeout_ms=None, **kwargs):
        raise NotImplementedError("implement in derived classes")

    def receive(self, size=None, timeout_ms=None, **kwargs):
        raise NotImplementedError("implement in derived classes")

    def query(self, msg_type, payload, timeout_ms=None, **kwargs):
        raise NotImplementedError("implement in derived classes")


class OOIProtocol(ProtocolInterface):

    msgs = {
        code: functools.partial(struct.Struct(msg).pack, code)
        for code, msg in {
            0x01: "<B",  # OP_INITIALIZE
            0x02: "<BI",  # OP_ITIME
            0x03: "<BH",  # set Strobe/Lamp enable Line
            0x05: "<BB",  # OP_GETINFO
            0x09: "<B",  # OP_REQUESTSPEC
            0x0A: "<BH",  # OP_SETTRIGMODE
            0x6A: "<BBH",  # OP_WRITE_REGISTER
            0x6B: "<BB",  # OP_READ_REGISTER
            0x71: "<BBB",  # OP_TECENABLE_QE
            0x72: "<B",  # OP_READTEC_QE
            0x73: "<Bh",  # OP_TECSETTEMP_QE
            0xFE: "<B",  # OP_USBMODE
        }.items()
    }  # add more here if you implement new features

    def __init__(self, transport):
        super(OOIProtocol, self).__init__(transport)
        # initialize the spectrometer
        self.send(0x01)
        time.sleep(0.1)  # wait shortly after init command

    def send(self, msg_type, payload=(), timeout_ms=None, **kwargs):
        """send a ooi message to the spectrometer

        Parameters
        ----------
        msg_type : int
            a message type as defined in `OOIProtocol.msgs`
        payload : optional
            dependent on `msg_type`. a singe value or a tuple of multiple values
        timeout_ms : int, optional
            the timeout after which the transport layer should error.
            `None` means no timeout (default)
        **kwargs :
            ignored and only present to provide compatible caller interfaces

        Returns
        -------
        bytes_written : int
            the number of bytes sent
        """
        if kwargs:
            warnings.warn("kwargs provided but ignored: {}".format(kwargs))
        payload = payload if isinstance(payload, (tuple, list)) else (payload,)
        data = self.msgs[msg_type](*payload)
        return self.transport.write(data, timeout_ms=timeout_ms)

    def receive(self, size=None, timeout_ms=None, mode=None, **kwargs):
        """receive data from the spectrometer

        Parameters
        ----------
        size : int, optional
            number of bytes to receive. if `None` (default) uses the
            default size as specified in the transport layer.
        timeout_ms : int, optional
            the timeout after which the transport layer should error.
            `None` means no timeout (default)
        mode : str, optional
            transport layers can support different modes
            (i.e. {'low_speed', 'high_speed', 'high_speed_alt'} in the usb case)
        kwargs :
            ignored and only present to provide compatible caller interfaces

        Returns
        -------
        data : str
            data returned from the spectrometer
        """
        if kwargs:
            warnings.warn("kwargs provided but ignored: {}".format(kwargs))
        return self.transport.read(
            size=size, timeout_ms=timeout_ms, mode=mode, **kwargs
        )

    def query(self, msg_type, payload, size=None, timeout_ms=None, mode=None, **kwargs):
        """convenience method combining send and receive

        Parameters
        ----------
        msg_type : int
            a message type as defined in `OOIProtocol.msgs`
        payload :
            the payload to be sent. Can be a singe value or a tuple, dependent
            `msg_type`.
        size : int, optional
            number of bytes to receive. if `None` (default) uses the
            default size as specified in the transport layer.
        timeout_ms : int, optional
            the timeout after which the transport layer should error.
            `None` means no timeout (default)
        mode : str, optional
            transport layers can support different modes
            (i.e. {'low_speed', 'high_speed', 'high_speed_alt'} in the usb case)
        kwargs :
            ignored and only present to provide compatible caller interfaces

        Returns
        -------
        data : str
            data returned from the spectrometer
        """
        self.send(msg_type, payload, timeout_ms=timeout_ms)
        return self.receive(size=size, timeout_ms=timeout_ms, **kwargs)


class OBPProtocol(ProtocolInterface):

    msgs = {
        code: struct.Struct(msg).pack
        for code, msg in {
            0x00000100: "",  # GET_SERIAL
            0x00000101: "",  # GET_SERIAL_LENGTH
            0x00100928: "",  # GET_BUF_SPEC32_META
            0x00101000: "",  # GET_RAW_SPECTRUM_NOW_HDX
            0x00101100: "",  # GET_RAW_SPECTRUM_NOW
            0x00110010: "<L",  # SET_ITIME_USEC
            0x00110110: "<B",  # SET_TRIG_MODE
            0x00180100: "",  # GET_WL_COEFF_COUNT
            0x00180101: "<B",  # GET_WL_COEFF
            0x00181100: "",  # GET_NL_COEFF_COUNT
            0x00181101: "<B",  # GET_NL_COEFF
            0x00420004: "",  # GET_TE_TEMPERATURE
            0x00420010: "<B",  # SET_TE_ENABLE
            0x00420011: "<f",  # SET_TE_SETPOINT
        }.items()
    }  # add more here if you implement new features

    class OBP(object):
        """All relevant constants are stored here"""

        HEADER_START_BYTES = 0xC0C1
        HEADER_PROTOCOL_VERSION = 0x1100  # XXX: seems to be the newest protocol version

        FLAG_RESPONSE_TO_REQUEST = 0x0001
        FLAG_ACK = 0x0002
        FLAG_REQUEST_ACK = 0x0004
        FLAG_NACK = 0x0008
        FLAG_HW_EXCEPTION = 0x0010
        FLAG_PROTOCOL_DEPRECATED = 0x0020

        ERROR_CODES = {
            0: "Success (no detectable errors)",
            1: "Invalid/unsupported protocol",
            2: "Unknown message type",
            3: "Bad checksum",
            4: "Message too large",
            5: "Payload length does not match message type",
            6: "Payload data invalid",
            7: "Device not ready for given message type",
            8: "Unknown checksum type",
            9: "Device reset unexpectedly",
            10: "Too many buses (Commands have come from too many bus interfaces)",
            11: "Out of memory. Failed to allocate enough space to complete request.",
            12: "Command is valid, but desired information does not exist.",
            13: "Int Device Error. May be unrecoverable.",
            100: "Could not decrypt properly",
            101: "Firmware layout invalid",
            102: "Data packet was wrong size",
            103: "hardware revision not compatible with firmware",
            104: "Existing flash map not compatible with firmware",
            255: "Operation/Response Deferred. Will take some time. Do not ACK or NACK yet.",
        }

        NO_ERROR = 0x0000
        RESERVED = b""
        CHECKSUM_TYPE_NONE = 0x00
        CHECKSUM_TYPE_MD5 = 0x01
        NO_CHECKSUM = b""
        FOOTER = 0xC2C3C4C5  # the datasheet specifies it in this order...

        HEADER_FMT = (
            "<H"  # start_bytes
            "H"  # protocol_version
            "H"  # flags
            "H"  # error number
            "L"  # message type
            "L"  # regarding
            "6s"  # reserved
            "B"  # checksum type
            "B"  # immediate length
            "16s"  # immediate data
            "L"  # bytes remaining
        )

        FOOTER_FMT = "16s" "L"  # checksum  # footer

    # noinspection DuplicatedCode
    def send(self, msg_type, payload=(), timeout_ms=None, request_ack=True, **kwargs):
        """send a obp message to the spectrometer

        Parameters
        ----------
        msg_type : int
            a message type as defined in `OBPProtocol.msgs`
        payload : optional
            dependent on `msg_type`. Can be a single value or a tuple of multiple values
        timeout_ms : int, optional
            the timeout after which the transport layer should error.
            `None` means no timeout (default)
        request_ack : bool, default `True`
            request an ack for the sent command from the spectrometer.
        **kwargs :
            ignored and only present to provide compatible caller interfaces

        Returns
        -------
        bytes_written : int
            the number of bytes sent
        """
        if kwargs:
            warnings.warn("kwargs provided but ignored: {}".format(kwargs))
        payload = payload if isinstance(payload, (tuple, list)) else (payload,)
        data = self.msgs[msg_type](*payload)

        # Constructing message and querying usb.
        message = self._construct_outgoing_message(
            msg_type, data, request_ack=request_ack
        )  # TODO
        bytes_written = self.transport.write(
            message
        )  # ? assert bytes_written == len(message)

        if not request_ack:
            return bytes_written

        response = self.transport.read(timeout_ms=timeout_ms)
        try:
            remaining_bytes, checksum_type = self._check_incoming_message_header(
                response[:44]
            )
        except SeaBreezeError:
            return 0
        # ? assert remaining_bytes == 20
        checksum = self._check_incoming_message_footer(response[-20:])

        if (
            checksum_type == self.OBP.CHECKSUM_TYPE_MD5
            and checksum != hashlib.md5(response[:-20]).digest()
        ):
            warnings.warn(
                "WARNING OBP: The checksums differ, but we ignore this for now."
            )
        return bytes_written

    def receive(self, size=None, timeout_ms=None, **kwargs):
        """receive data from the spectrometer

        Parameters
        ----------
        size : int, optional
            number of bytes to receive. if `None` (default) uses the
            default size as specified in the transport layer.
        timeout_ms : int, optional
            the timeout after which the transport layer should error.
            `None` means no timeout (default)
        kwargs :
            ignored and only present to provide compatible caller interfaces

        Returns
        -------
        data : str
            data returned from the spectrometer
        """
        response = self.transport.read(size=64, timeout_ms=timeout_ms)
        try:
            remaining_bytes, checksum_type = self._check_incoming_message_header(
                response[:44]
            )
        except SeaBreezeError:
            # empty buffer if error raised
            self.transport.read(size=None, timeout_ms=500)
            raise
        length_payload_footer = remaining_bytes

        # we already received some data
        remaining_bytes -= len(response[44:])
        if remaining_bytes > 0:
            response += self.transport.read(size=remaining_bytes, timeout_ms=timeout_ms)

        if length_payload_footer != len(response[44:]):
            raise SeaBreezeError(
                "remaining packet length mismatch: {:d} != {:d}".format(
                    remaining_bytes, len(response[44:])
                )
            )

        checksum = self._check_incoming_message_footer(response[-20:])
        if (
            checksum_type == self.OBP.CHECKSUM_TYPE_MD5
            and checksum != hashlib.md5(response[:-20]).digest()
        ):
            warnings.warn(
                "WARNING OBP: The checksums differ, but we ignore this for now."
            )

        return self._extract_message_data(response)

    def query(self, msg_type, payload=(), size=None, timeout_ms=None, **kwargs):
        """convenience method combining send and receive

        Parameters
        ----------
        msg_type : int
            a message type as defined in `OBPProtocol.msgs`
        payload :
            the payload to be sent. Can be a single value or a tuple, dependent
            `msg_type`.
        size : int, optional
            number of bytes to receive. if `None` (default) uses the
            default size as specified in the transport layer.
        timeout_ms : int, optional
            the timeout after which the transport layer should error.
            `None` means no timeout (default)
        kwargs :
            ignored and only present to provide compatible caller interfaces

        Returns
        -------
        data : str
            data returned from the spectrometer
        """
        self.send(msg_type, payload, request_ack=False)
        return self.receive(timeout_ms=timeout_ms)

    def _construct_outgoing_message(
        self, msg_type, payload_string, request_ack=False, regarding=None
    ):
        """construct an outgoing OBP message

        Parameters
        ----------
        msg_type : `int`
            the obp message type, a 4 byte integer
        payload_string : `bytes`
            a compiled payload_string
        request_ack : `bool`
            request an ack for the sent command from the spectrometer.
        regarding : `int`
            see ocean optics obp protocol documentation

        Returns
        -------
        message : `str`
            compiled message
        """
        if request_ack is True:
            flags = self.OBP.FLAG_REQUEST_ACK
        else:
            flags = 0

        if regarding is None:
            regarding = 0

        if len(payload_string) <= 16:
            payload_string_fmt = "0s"
            immediate_length = len(payload_string)
            immediate_data = payload_string
            payload_string = b""
            bytes_remaining = 20  # Checksum + footer
        else:
            payload_string_fmt = "%ds" % len(payload_string)
            immediate_length = 0
            immediate_data = b""
            bytes_remaining = 20 + len(payload_string)

        FMT = self.OBP.HEADER_FMT + payload_string_fmt + self.OBP.FOOTER_FMT

        msg = struct.pack(
            FMT,
            self.OBP.HEADER_START_BYTES,
            self.OBP.HEADER_PROTOCOL_VERSION,
            flags,
            self.OBP.NO_ERROR,
            msg_type,
            regarding,
            self.OBP.RESERVED,
            self.OBP.CHECKSUM_TYPE_NONE,
            immediate_length,
            immediate_data,
            bytes_remaining,
            payload_string,
            self.OBP.NO_CHECKSUM,
            self.OBP.FOOTER,
        )
        return msg

    def _check_incoming_message_header(self, header):
        """check the incoming message header

        Parameters
        ----------
        header : `str`
            a obp header of length 44

        Returns
        -------
        bytes_and_checksum_type : tuple[`int`, `int`]
            bytes_remaining after the header (returns 20 for a 64 byte message)
            checksum_type only supports self.OBP.CHECKSUM_TYPE_MD5 for now
        """
        if len(header) != 44:
            raise SeaBreezeError(
                "header has wrong length! len(header): %d" % len(header)
            )

        data = struct.unpack(self.OBP.HEADER_FMT, header)

        if data[0] != self.OBP.HEADER_START_BYTES:
            raise SeaBreezeError('Header start_bytes wrong: "%d"' % data[0])
        if data[1] != self.OBP.HEADER_PROTOCOL_VERSION:
            raise SeaBreezeError("Header protocol version wrong: %d" % data[1])

        flags = data[2]
        if flags == 0:
            pass
        if flags & self.OBP.FLAG_RESPONSE_TO_REQUEST:
            pass  # TODO: propagate?
        if flags & self.OBP.FLAG_ACK:
            pass  # TODO: propagate?
        if flags & self.OBP.FLAG_REQUEST_ACK:
            pass  # TODO: only the host should be able to set this?
        if (flags & self.OBP.FLAG_NACK) or (flags & self.OBP.FLAG_HW_EXCEPTION):
            error = data[3]
            if error != 0:  # != SUCCESS
                raise SeaBreezeError(self.OBP.ERROR_CODES[error])
            else:
                pass  # TODO: should we do something here?
        if flags & self.OBP.FLAG_PROTOCOL_DEPRECATED:
            raise SeaBreezeError("Protocol deprecated?!?")

        # msg_type = data[4]
        # regarding = data[5]

        checksum_type = data[7]  # TODO: implement checksums.
        if checksum_type not in [
            self.OBP.CHECKSUM_TYPE_NONE,
            self.OBP.CHECKSUM_TYPE_MD5,
        ]:
            raise SeaBreezeError('the checksum type is unknown: "%d"' % checksum_type)

        # immediate_length = data[8]
        # immediate_data = data[9]
        bytes_remaining = data[10]

        return bytes_remaining, checksum_type

    def _check_incoming_message_footer(self, footer):
        """check the incoming message header

        Parameters
        ----------
        footer : `str`
            a obp footer of length 20

        Returns
        -------
        checksum: `str`
            the 16 byte checksum of the message
        """
        assert len(footer) == 20, "footer has wrong length! len(footer): %d" % len(
            footer
        )

        data = struct.unpack("<" + self.OBP.FOOTER_FMT, footer)

        checksum = data[0]
        assert data[1] == self.OBP.FOOTER, (
            "the device returned a wrong footer: %d" % data[1]
        )

        return checksum

    def _extract_message_data(self, msg):
        """extract the payload data from a obp message

        Parameters
        ----------
        msg : `str`
            a obp message

        Returns
        -------
        data : `str`
            the payload contained in the message
        """
        payload_length = len(msg) - 44 - 20  # - HeaderLength - FooterLength
        if not (payload_length >= 0):
            raise SeaBreezeError("Received message < 64 bytes: %d" % payload_length)
        payload_fmt = "%ds" % payload_length
        FMT = self.OBP.HEADER_FMT + payload_fmt + self.OBP.FOOTER_FMT

        data = struct.unpack(FMT, msg)

        msg_type = data[4]

        immediate_length = data[8]
        immediate_data = data[9]
        payload = data[11]

        if (immediate_length > 0) and len(payload) > 0:
            raise SeaBreezeError("Got immediate AND payload data? cmd: '%d'" % msg_type)
        elif immediate_length > 0:
            return immediate_data[:immediate_length]
        elif payload_length > 0:
            return payload
        else:
            return b""



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezePixelBinningFeature(SeaBreezeFeature):
    identifier = "pixel_binning"

    def set_binning_factor(self, factor):
        raise NotImplementedError("implement in derived class")

    def get_binning_factor(self):
        raise NotImplementedError("implement in derived class")

    def get_max_binning_factor(self):
        raise NotImplementedError("implement in derived class")

    binning_factor = property(get_binning_factor, set_binning_factor)
    max_binning_factor = property(get_max_binning_factor)

    def set_default_binning_factor(self, factor):
        raise NotImplementedError("implement in derived class")

    def get_default_binning_factor(self):
        raise NotImplementedError("implement in derived class")

    default_binning_factor = property(
        get_default_binning_factor, set_default_binning_factor
    )



"""Implementation of the Seabreeze Transport layer.

Some spectrometers can support different transports (usb, network, rs232, etc.)

"""
import importlib
import inspect
import logging
import warnings
from functools import partial

import usb.core
import usb.util

try:
    from functools import partialmethod
except ImportError:
    # https://gist.github.com/carymrobbins/8940382
    # noinspection PyPep8Naming
    class partialmethod(partial):
        def __get__(self, instance, owner):
            if instance is None:
                return self
            args, kwargs = self.args or (), self.keywords or {}
            return partial(self.func, instance, *args, **kwargs)


class TransportInterface(object):

    _required_init_kwargs = ()

    def open_device(self, device):
        """open a seabreeze device

        Parameters
        ----------
        device : seabreeze.pyseabreeze.devices.SeaBreezeDevice
        """
        raise NotImplementedError("implement in derived transport class")

    @property
    def is_open(self):
        """return if device is opened

        Returns
        -------
        bool
        """
        raise NotImplementedError("implement in derived transport class")

    def close_device(self):
        """close the seabreeze device"""
        raise NotImplementedError("implement in derived transport class")

    def write(self, data, timeout_ms=None, **kwargs):
        """write data to the device"""
        raise NotImplementedError("implement in derived transport class")

    def read(self, size=None, timeout_ms=None, **kwargs):
        """read data from the

        Returns
        -------
        str
        """
        raise NotImplementedError("implement in derived transport class")

    @property
    def default_timeout_ms(self):
        raise NotImplementedError("implement in derived transport class")

    @property
    def protocol(self):
        raise NotImplementedError("implement in derived transport class")

    @classmethod
    def list_devices(cls):
        raise NotImplementedError("implement in derived transport class")

    @classmethod
    def register_model(cls, model_name, **kwargs):
        raise NotImplementedError("implement in derived transport class")

    @classmethod
    def supported_model(cls, device):
        """return if the device supports the transport or vice versa

        Returns
        -------
        model_name : ``
        """
        return None

    @classmethod
    def specialize(cls, model_name, **kwargs):
        raise NotImplementedError("implement in derived transport class")


# encapsulate usb.core.USBError
class USBTransportError(Exception):
    def __init__(self, *args, **kwargs):
        super(USBTransportError, self).__init__(*args)
        self.errno = kwargs.pop("errno", None)
        self.backend_error_code = kwargs.pop("error_code", None)
        if kwargs:
            raise TypeError("USBTransportError got unexpected kwarg")

    @classmethod
    def from_usberror(cls, err):
        return cls(str(err), errno=err.errno, error_code=err.backend_error_code)


class USBTransportDeviceInUse(Exception):
    pass


# this can and should be opaque to pyseabreeze
class USBTransportHandle(object):
    def __init__(self, pyusb_device):
        """encapsulation for pyusb device classes

        Parameters
        ----------
        pyusb_device : usb.core.Device
        """
        self.pyusb_device = pyusb_device
        # noinspection PyUnresolvedReferences
        self.identity = (
            pyusb_device.idVendor,
            pyusb_device.idProduct,
            pyusb_device.bus,
            pyusb_device.address,
        )
        self.pyusb_backend = get_name_from_pyusb_backend(pyusb_device.backend)

    def close(self):
        try:
            self.pyusb_device.reset()
        except usb.core.USBError:
            logging.debug(
                "USBError while calling USBTransportHandle.close on {:04x}:{:04x}".format(
                    self.identity[0], self.identity[1]
                ),
                exc_info=True,
            )

    def __del__(self):
        if self.pyusb_backend == "libusb1":
            # have to check if .finalize() has been called
            # -> todo: maybe better to fix this in the api initialization of cseabreeze
            # -> todo: will probably have to check pyusb versions and only do this when necessary
            if not getattr(self.pyusb_device.backend, "_finalize_called", False):
                # if usb.core.Device.reset() gets called but the backend has been finalized already
                # (this happens only during interpreter shutdown)
                self.close()
        else:
            self.close()
        self.pyusb_device = None


class USBTransport(TransportInterface):
    """implementation of the usb transport interface for spectrometers"""

    _required_init_kwargs = ("usb_product_id", "usb_endpoint_map", "usb_protocol")
    vendor_id = 0x2457
    product_ids = {}

    # add logging
    _log = logging.getLogger(__name__)

    def __init__(self, usb_product_id, usb_endpoint_map, usb_protocol):
        super(USBTransport, self).__init__()
        self._product_id = usb_product_id
        self._endpoint_map = usb_endpoint_map
        self._protocol_cls = usb_protocol
        # internal settings
        self._default_read_size = {
            "low_speed": 64,
            "high_speed": 512,
            "high_speed_alt": 512,
        }
        self._read_endpoints = {
            "low_speed": "lowspeed_in",
            "high_speed": "highspeed_in",
            "high_speed_alt": "highspeed_in2",
        }
        self._default_read_endpoint = "low_speed"
        self._default_read_spectrum_endpoint = "high_speed"
        # internal state
        self._device = None
        self._opened = None
        self._protocol = None

    def open_device(self, device):
        if not isinstance(device, USBTransportHandle):
            raise TypeError("device needs to be a USBTransportHandle")
        # device.reset()
        self._device = device
        pyusb_device = self._device.pyusb_device
        try:
            if pyusb_device.is_kernel_driver_active(0):
                pyusb_device.detach_kernel_driver(0)
        except NotImplementedError:
            pass  # unavailable on some systems/backends
        try:
            pyusb_device.set_configuration()
        except usb.core.USBError as err:
            if err.errno == 16:
                # TODO: warn as in cseabreeze
                self._opened = True
                raise USBTransportDeviceInUse(
                    "device probably used by another thread/process"
                )
            raise USBTransportError.from_usberror(err)
        else:
            self._opened = True
        # This will initialize the communication protocol
        if self._opened:
            self._protocol = self._protocol_cls(self)

    @property
    def is_open(self):
        return self._opened or False

    def close_device(self):
        if self._device is not None:
            self._device.close()
            self._device = None
        self._opened = False
        self._protocol = None

    def write(self, data, timeout_ms=None, **kwargs):
        if self._device is None:
            raise RuntimeError("device not opened")
        if kwargs:
            warnings.warn("kwargs provided but ignored: {}".format(kwargs))
        return self._device.pyusb_device.write(
            self._endpoint_map.ep_out, data, timeout=timeout_ms
        )

    def read(self, size=None, timeout_ms=None, mode=None, **kwargs):
        if self._device is None:
            raise RuntimeError("device not opened")
        mode = mode if mode is not None else self._default_read_endpoint
        endpoint = getattr(self._endpoint_map, self._read_endpoints[mode])
        if size is None:
            size = self._default_read_size[mode]
        if kwargs:
            warnings.warn("kwargs provided but ignored: {}".format(kwargs))
        return self._device.pyusb_device.read(endpoint, size, timeout=timeout_ms)

    @property
    def default_timeout_ms(self):
        return self._device.pyusb_device.default_timeout if self._device else None

    @property
    def protocol(self):
        return self._protocol

    @classmethod
    def list_devices(cls, **kwargs):
        """list pyusb devices for all available spectrometers

        Note: this includes spectrometers that are currently opened in other
        processes on the machine.

        Yields
        ------
        devices : USBTransportHandle
            unique pyusb devices for each available spectrometer
        """
        # check if a specific pyusb backend is requested
        _pyusb_backend = kwargs.get("_pyusb_backend", None)
        # get all matching devices
        try:
            pyusb_devices = usb.core.find(
                find_all=True,
                custom_match=lambda dev: (
                    dev.idVendor == cls.vendor_id and dev.idProduct in cls.product_ids
                ),
                backend=get_pyusb_backend_from_name(name=_pyusb_backend),
            )
        except usb.core.NoBackendError:
            raise RuntimeError("No pyusb backend found")
        # encapsulate
        for pyusb_device in pyusb_devices:
            yield USBTransportHandle(pyusb_device)

    @classmethod
    def register_model(cls, model_name, **kwargs):
        product_id = kwargs.get("usb_product_id")
        if product_id in cls.product_ids:
            raise ValueError(
                "product_id 0x{:04x} already in registry".format(product_id)
            )
        cls.product_ids[product_id] = model_name

    @classmethod
    def supported_model(cls, device):
        """return supported model

        Parameters
        ----------
        device : USBTransportHandle
        """
        if not isinstance(device, USBTransportHandle):
            return None
        # noinspection PyUnresolvedReferences
        return cls.product_ids[device.pyusb_device.idProduct]

    @classmethod
    def specialize(cls, model_name, **kwargs):
        assert set(kwargs) == set(cls._required_init_kwargs)
        # usb transport register automatically on registration
        cls.register_model(model_name, **kwargs)
        specialized_class = type(
            "USBTransport{}".format(model_name),
            (cls,),
            {"__init__": partialmethod(cls.__init__, **kwargs)},
        )
        return specialized_class

    @classmethod
    def initialize(cls, **_kwargs):
        for device in cls.list_devices(**_kwargs):
            try:
                device.pyusb_device.reset()
                # usb.util.dispose_resources(device)  <- already done by device.reset()
            except Exception as err:
                cls._log.debug(
                    "initialize failed: {}('{}')".format(
                        err.__class__.__name__, getattr(err, "message", "no message")
                    )
                )

    @classmethod
    def shutdown(cls, **_kwargs):
        # dispose usb resources
        for device in cls.list_devices(**_kwargs):
            try:
                usb.util.dispose_resources(device.pyusb_device)
            except Exception as err:
                cls._log.debug(
                    "shutdown failed: {}('{}')".format(
                        err.__class__.__name__, getattr(err, "message", "no message")
                    )
                )


_pyusb_backend_instances = {}


def get_pyusb_backend_from_name(name):
    """internal: allow requesting a specific pyusb backend for testing"""
    if name is None:
        # default is pick first that works: ('libusb1', 'libusb0', 'openusb0')
        _backend = None
    else:
        try:
            _backend = _pyusb_backend_instances[name]
        except KeyError:
            m = importlib.import_module("usb.backend.{}".format(name))
            # noinspection PyUnresolvedReferences
            _backend = m.get_backend()
            # raise if a pyusb backend was requested but can't be loaded
            if _backend is None:
                raise RuntimeError("pyusb '{}' backend failed to load")
            _pyusb_backend_instances[name] = _backend
    return _backend


def get_name_from_pyusb_backend(backend):
    """internal: return backend name from loaded backend"""
    module = inspect.getmodule(backend)
    if not module:
        return None
    return module.__name__.split(".")[-1]



"""This is the pyusb implementation of the seabreeze library

Author: Andreas Poehlmann

most of this code is based on implementations from
    http://github.com/ap--/python-oceanoptics

"""
import logging
import weakref

from seabreeze.pyseabreeze.devices import SeaBreezeDevice, _model_class_registry
from seabreeze.pyseabreeze.transport import (
    USBTransport,
    USBTransportDeviceInUse,
    USBTransportError,
    USBTransportHandle,
)

__all__ = ["SeaBreezeAPI"]

# create only one SeaBreezeDevice instance per handle
_seabreeze_device_instance_registry = weakref.WeakValueDictionary()


def _seabreeze_device_factory(device):
    """return existing instances instead of creating temporary ones

    Parameters
    ----------
    device : USBTransportHandle

    Returns
    -------
    dev : SeaBreezeDevice
    """
    global _seabreeze_device_instance_registry
    if not isinstance(device, USBTransportHandle):
        raise TypeError("needs to be instance of USBTransportHandle")
    ident = device.identity
    try:
        return _seabreeze_device_instance_registry[ident]
    except KeyError:
        dev = _seabreeze_device_instance_registry[ident] = SeaBreezeDevice(device)
        return dev


class SeaBreezeAPI(object):
    """SeaBreeze API interface"""

    _log = logging.getLogger(__name__)

    def __init__(self, initialize=True, **_kwargs):
        self._kwargs = _kwargs  # allow passing additional kwargs to transports
        if initialize:
            self.initialize()

    def initialize(self):
        """initialize the api backend

        normally this function does not have to be called directly by the user.
        it resets all usb devices on load
        """
        USBTransport.initialize(**self._kwargs)

    def shutdown(self):
        """shutdown the api backend

        normally this function does not have to be called directly by the user
        """
        # dispose usb resources
        _seabreeze_device_instance_registry.clear()
        USBTransport.shutdown(**self._kwargs)

    def add_rs232_device_location(self, device_type, bus_path, baudrate):
        """add RS232 device location"""
        # RS232Transport.register_device(device_type, bus_path, baudrate)
        raise NotImplementedError("rs232 communication not implemented for pyseabreeze")

    def add_ipv4_device_location(self, device_type, ip_address, port):
        """add ipv4 device location"""
        # IPV4Transport.register_device(device_type, ip_address, port)
        raise NotImplementedError("ipv4 communication not implemented for pyseabreeze")

    def list_devices(self):
        """returns available SeaBreezeDevices

        list all connected Ocean Optics devices supported
        by libseabreeze.

        Returns
        -------
        devices: list of SeaBreezeDevice
            connected Spectrometer instances
        """
        # get all matching devices
        devices = []
        for usb_dev in USBTransport.list_devices(**self._kwargs):
            # get the correct communication interface
            dev = _seabreeze_device_factory(usb_dev)
            if not dev.is_open:
                # opening the device will populate its serial number
                try:
                    dev.open()
                except USBTransportDeviceInUse:
                    # device used by another thread? -> exclude
                    continue
                except USBTransportError:
                    # todo: investigate if there are ways to recover from here
                    raise
                else:
                    dev.close()
            devices.append(dev)
        return devices

    # note: to be fully consistent with cseabreeze this shouldn't be a staticmethod
    @staticmethod
    def supported_models():
        """returns SeaBreezeDevices supported by the backend

        models supported by the backend

        Returns
        -------
        devices: list of str
            list of model names that are supported by this backend
        """
        return [
            x for x in sorted(_model_class_registry.keys()) if not x.startswith("_")
        ]



import logging
import sys
import warnings

__all__ = ["use", "get_backend"]

_log = logging.getLogger(__name__)


def _use_backend(name):
    # internal: import the libseabreeze cython wrapper -> cseabreeze
    try:
        if name == "cseabreeze":
            import seabreeze.cseabreeze as sbb
        elif name == "pyseabreeze":
            import seabreeze.pyseabreeze as sbb
        else:
            raise ValueError("unknown backend '{}'".format(name))
    except ImportError as err:
        _log.warning(
            "seabreeze can't load '{}' backend - error: '{}'".format(name, repr(err)),
            exc_info=True,
        )
        return None
    else:
        return sbb


# internal config dict used exclusively in this submodule
_SeaBreezeConfig = {
    "requested_backend": "cseabreeze",  # default is cseabreeze
    "available_backends": ("cseabreeze", "pyseabreeze"),
    "allow_fallback": False,
    "_api_kwargs": {},  # for pytests
}


def use(backend, force=True, **_kwargs):
    """
    select the backend used for communicating with the spectrometer

    Parameters
    ----------
    backend : str

    force : bool, optional, default: True
        raises an ImportError when ``seabreeze.get_backend()`` is called
        and the requested backend can not be imported. force=True should
        be used in user code to ensure that the correct backend is being
        loaded.
    """
    if backend not in _SeaBreezeConfig["available_backends"]:
        raise ValueError(
            "backend not in: {}".format(
                ", ".join(_SeaBreezeConfig["available_backends"])
            )
        )
    if 'seabreeze.spectrometers' in sys.modules:
        warnings.warn(
            "seabreeze.use has to be called before importing seabreeze.spectrometers",
        )

    _SeaBreezeConfig["requested_backend"] = backend
    _SeaBreezeConfig["allow_fallback"] = not force
    _SeaBreezeConfig["_api_kwargs"] = _kwargs.pop("_api_kwargs", {})
    if _kwargs:
        raise TypeError("unknown keyword arguments")


def get_backend():
    """
    return the requested backend or a fallback. configuration is done
    via ``seabreeze.use()``

    Returns
    -------
    backend:
        a backend interface for communicating with the spectrometers
    """
    requested = _SeaBreezeConfig["requested_backend"]
    fallback = _SeaBreezeConfig["allow_fallback"]
    backends = _SeaBreezeConfig["available_backends"]

    backend = _use_backend(requested)  # trying to import requested backend
    if backend is None and fallback:

        warnings.warn(
            "seabreeze failed to load user requested '{}' backend but will try fallback backends".format(
                requested
            ),
            category=ImportWarning,
        )

        fallback_backends = [b for b in backends if b != requested]
        for fallback_backend in fallback_backends:
            backend = _use_backend(fallback_backend)
            if backend is not None:
                break
        else:
            raise ImportError(
                "Could not import backend. Tried: {}".format(", ".join(backends))
            )

    if backend is None:
        raise ImportError("Could not import backend. Requested: {}".format(requested))

    # provide for testing purposes
    backend._api_kwargs = _SeaBreezeConfig["_api_kwargs"]

    return backend



"""operating system relevant deploy functions

author: Andreas Poehlmann
"""
import argparse
import ctypes
import logging
import os
import platform
import shutil
import subprocess
import sys
import tempfile
import time
import zipfile
from builtins import input
from textwrap import dedent

from future.standard_library import hooks

with hooks():
    from urllib.error import HTTPError
    from urllib.request import urlopen

try:
    # noinspection PyProtectedMember
    from textwrap import indent as _indent
except ImportError:
    # noinspection PyUnusedLocal
    def _indent(text, prefix, predicate=None):
        return u"".join(prefix + line for line in text.splitlines(True))


_GITHUB_REPO_URL = (
    "https://raw.githubusercontent.com/ap--/python-seabreeze/master/os_support"
)
_UDEV_RULES_PATH = "/etc/udev/rules.d/10-oceanoptics.rules"
_DRIVERS_ZIP_FN = "windows-driver-files.zip"
_log = logging.getLogger(__name__)


def _diff_files(file1, file2):
    """diff two files using linux `diff`"""
    try:
        return subprocess.check_output(["diff", file1, file2]).decode("utf8")
    except subprocess.CalledProcessError as err:
        return err.output.decode("utf8")


def _request_confirmation(question):
    """require user input to continue"""
    while True:
        user_input = input("{} [y/n] ".format(question)).lower()
        if user_input not in {"y", "n"}:
            _log.info("Please enter 'y' or 'n'.")
        elif user_input[0] == "n":
            return False
        else:
            return True
        time.sleep(0.1)


def linux_install_udev_rules():
    """verify and install the udev rules"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--overwrite-existing",
        help="overwrite rules if already present",
        action="store_true",
    )
    parser.add_argument(
        "rules",
        help="rules file (default: download from github)",
        default="",
        nargs="?",
    )
    args = parser.parse_args()

    if args.rules:
        if not os.path.exists(args.rules):
            raise IOError("rules file '{}' doesn't exist".format(args.rules))
        udev_tmp_file = None
        udev_fn = args.rules

    else:
        udev_tmp_file = tempfile.NamedTemporaryFile()
        udev_fn = udev_tmp_file.name

    try:
        # download rules from github if no file is provided
        if udev_tmp_file is not None:
            url = "{}/{}".format(_GITHUB_REPO_URL, os.path.basename(_UDEV_RULES_PATH))
            try:
                _log.info("downloading rules from github")
                udev_data = urlopen(url).read()
            except HTTPError:
                _log.error("can't download '{}'".format(url))
                sys.exit(1)
            udev_tmp_file.write(udev_data)
            udev_tmp_file.flush()

        # check if rules need to be overwritten
        if os.path.exists(_UDEV_RULES_PATH) and not args.overwrite_existing:

            rules_differ = _diff_files(_UDEV_RULES_PATH, udev_fn)
            if not rules_differ:
                _log.info("udev rules already newest version")
                sys.exit(0)
            else:
                _log.info(_indent(rules_differ, u"  ").rstrip())
                _log.info(
                    "udev rules differ. To overwrite run with '--overwrite-existing'"
                )
                sys.exit(1)

        if not _request_confirmation("Install udev rules?"):
            sys.exit(0)

        # cp rules and execute
        _log.info("Copying udev rules to {}".format(_UDEV_RULES_PATH))
        subprocess.call(["sudo", "cp", udev_fn, _UDEV_RULES_PATH])
        _log.info("Calling udevadm control --reload-rules")
        subprocess.call(["sudo", "udevadm", "control", "--reload-rules"])
        _log.info("Success")
        sys.exit(0)

    finally:
        if udev_tmp_file is not None:
            udev_tmp_file.close()  # removes tempfile


def _windows_is_admin():
    """windows only: check if running as admin"""
    # noinspection PyBroadException
    try:
        return ctypes.windll.shell32.IsUserAnAdmin()
    except Exception:
        return False


def _is_contained_in_dir(files, cdir=None):
    cdir = os.path.abspath(cdir or os.path.curdir)
    for f in files:
        f_abs = os.path.abspath(f)
        if not os.path.commonprefix((f_abs, cdir)).startswith(cdir):
            return False
    return True


def _unicode(x):
    try:
        # noinspection PyUnresolvedReferences
        return unicode(x)
    except NameError:
        return x


def windows_install_drivers():
    """install driver inf files via pnputil in an elevated shell"""
    if not _request_confirmation("Install windows drivers?"):
        sys.exit(0)

    if not _windows_is_admin():
        # Re-run the program with admin rights
        argv = [__file__] + sys.argv[1:]
        ret = ctypes.windll.shell32.ShellExecuteW(
            None,
            _unicode("runas"),
            _unicode(sys.executable),
            _unicode(subprocess.list2cmdline(argv)),
            None,
            1,
        )
        if ret > 32:
            _log.info("Launched admin shell")
        else:
            _log.info("Failed to launch admin shell. Error code {}".format(ret))
        sys.exit(0 if ret > 32 else 1)

    # running as admin
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "drivers_zip",
        help="drivers zip file (default: download from github)",
        default="",
        nargs="?",
    )
    args = parser.parse_args()

    if args.drivers_zip:
        if not os.path.exists(args.drivers_zip):
            raise IOError(
                "drivers_zip file '{}' doesn't exist".format(args.drivers_zip)
            )
        drivers_zip = args.drivers_zip
    else:
        drivers_zip = None

    tmp_dir = tempfile.mkdtemp(prefix="seabreeze-os-")
    # noinspection PyBroadException
    try:
        # download driver files
        if drivers_zip is None:
            url = "{}/{}".format(_GITHUB_REPO_URL, os.path.basename(_DRIVERS_ZIP_FN))
            drivers_zip = os.path.join(tmp_dir, _DRIVERS_ZIP_FN)
            with open(drivers_zip, "wb") as dzip:
                try:
                    _log.info("Downloading windows drivers from github")
                    drivers_zip_data = urlopen(url).read()
                except HTTPError:
                    _log.error("Can't download '{}'".format(url))
                    sys.exit(1)
                dzip.write(drivers_zip_data)

        # extract driver files
        with zipfile.ZipFile(drivers_zip, "r") as dzip:
            if not _is_contained_in_dir(dzip.namelist()):
                raise Exception("Zipfile contains non subdir paths")
            dzip.extractall(tmp_dir)
        _log.info("Extracted to temporary directory {}".format(tmp_dir))

        # use correct pnputil with 32bit pythons
        if "32bit" in platform.architecture():
            pnputil = r"%systemroot%\Sysnative\pnputil.exe"
        else:
            pnputil = "pnputil.exe"

        # install with pnp util
        cmd = [pnputil, "-i", "-a", os.path.join(tmp_dir, "*.inf")]
        return_code = subprocess.call(cmd, shell=True)

        _log.warning(
            dedent(
                """\
            Note: Some of the drivers currently don't have valid signatures.
            Look at the output above. If the spectrometer you want to use only
            provides an unsigned driver, you might have to install it manually.
            If you encounter this issue, please report it on github."""
            )
        )

        if return_code == 0:
            _log.info("Success")
        elif return_code == 3010:
            _log.info("Success! REBOOT REQUIRED!")
        else:
            _log.error("pnputil returned with {}".format(return_code))

    except Exception:
        _log.error("Error when installing drivers", exc_info=True)

    finally:
        shutil.rmtree(tmp_dir)
        input("Press [enter] to close.")


def main():
    logging.basicConfig(level=logging.INFO, format="%(message)s")
    system = platform.system()
    if system == "Windows":
        windows_install_drivers()
    elif system == "Linux":
        linux_install_udev_rules()
    else:
        _log.info("Nothing to do for system '{}'".format(system))
    sys.exit(0)


if __name__ == "__main__":
    # fix for windows entry_point shims, which are actually .exe files...
    main()
